// Maria structure component reference expression class -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "StructComponent.h"
#include "StructType.h"
#include "StructValue.h"
#include "Constant.h"
#include "Printer.h"

/** @file StructComponent.C
 * Structure component reference
 */

StructComponent::StructComponent (class Expression& structure,
				  card_t index) :
  myStructure (&structure), myIndex (index)
{
  assert (myStructure && myStructure->getType ()->getKind () == Type::tStruct);
  const class StructType* type =
    static_cast<const class StructType*>(myStructure->getType ());
  assert (myStructure->isBasic ());
  setType ((*type)[myIndex]);
}

StructComponent::~StructComponent ()
{
  myStructure->destroy ();
}

class Value*
StructComponent::do_eval (const class Valuation& valuation) const
{
  class Value* v = myStructure->eval (valuation);
  if (!v)
    return NULL;

  assert (v->getKind () == Value::vStruct);
  class Value* u = NULL;
  if (const class Value* w = (*static_cast<class StructValue*>(v))[myIndex])
    u = w->copy ();
  delete v;
  return u;
}

class Expression*
StructComponent::ground (const class Valuation& valuation,
			 class Transition* transition,
			 bool declare)
{
  class Expression* structure =
    myStructure->ground (valuation, transition, declare);
  if (!structure)
    return NULL;
  assert (valuation.isOK ());
  if (structure == myStructure) {
    structure->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class StructComponent (*structure, myIndex);
    if (!expr->isBasic ());
    else if (class Value* v = expr->eval (valuation)) {
      expr->destroy ();
      return new class Constant (*v);
    }
    else if (!valuation.isOKorVar ()) {
      expr->destroy ();
      return NULL;
    }

    valuation.clearErrors ();
    return expr;
  }
}

class Expression*
StructComponent::substitute (class Substitution& substitution)
{
  class Expression* structure = myStructure->substitute (substitution);
  if (structure == myStructure) {
    structure->destroy ();
    return copy ();
  }
  else
    return new class StructComponent (*structure, myIndex);
}

bool
StructComponent::depends (const class VariableSet& vars,
			  bool complement) const
{
  return myStructure->depends (vars, complement);
}

bool
StructComponent::forVariables (bool (*operation)
			       (const class Expression&,void*),
			       void* data) const
{
  return myStructure->forVariables (operation, data);
}

#ifdef EXPR_COMPILE
# include "CExpression.h"

void
StructComponent::compile (class CExpression& cexpr,
			  unsigned indent,
			  const char* lvalue,
			  const class VariableSet* vars) const
{
  char* rvalue = 0;
  class StringBuffer& out = cexpr.getOut ();
  if (cexpr.getVariable (*myStructure, rvalue))
    myStructure->compile (cexpr, indent, rvalue, vars);
  out.indent (indent);
  out.append (lvalue);
  out.append ("=");
  out.append (rvalue);
  out.append (".s");
  out.append (myIndex);
  out.append (";\n");
  delete[] rvalue;
}

#endif // EXPR_COMPILE

/** Determine whether an expression needs to be enclosed in parentheses
 * @param kind	kind of the expression
 * @return	whether parentheses are necessary
 */
inline static bool
needParentheses (enum Expression::Kind kind)
{
  switch (kind) {
  case Expression::eVariable:
  case Expression::eUndefined:
  case Expression::eStructComponent:
  case Expression::eUnionComponent:
  case Expression::eVectorIndex:
    return false;
  case Expression::eCardinality:
  case Expression::eUnionType:
  case Expression::eUnion:
  case Expression::eVector:
  case Expression::eBinop:
  case Expression::eBooleanBinop:
  case Expression::eNot:
  case Expression::eRelop:
  case Expression::eBuffer:
  case Expression::eBufferRemove:
  case Expression::eBufferWrite:
  case Expression::eBufferIndex:
  case Expression::eSet:
  case Expression::eTemporalBinop:
  case Expression::eTemporalUnop:
  case Expression::eMarking:
  case Expression::eTransitionQualifier:
  case Expression::ePlaceContents:
  case Expression::eSubmarking:
  case Expression::eMapping:
  case Expression::eEmptySet:
  case Expression::eVectorAssign:
  case Expression::eVectorShift:
    assert (false);
  case Expression::eUnop:
  case Expression::eConstant:
  case Expression::eStruct:
  case Expression::eStructAssign:
  case Expression::eIfThenElse:
  case Expression::eBufferUnop:
  case Expression::eTypecast:
    break;
  }

  return true;
}

void
StructComponent::display (const class Printer& printer) const
{
  if (::needParentheses (myStructure->getKind ())) {
    printer.delimiter ('(')++;
    myStructure->display (printer);
    --printer.delimiter (')');
  }
  else
    myStructure->display (printer);

  printer.delimiter ('.');
  printer.print (static_cast<const class StructType*>(myStructure->getType ())
		 ->getComponentName (myIndex));
}
